! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************

!==================================================================================
!
! This extension contains a simple, coherent presecription to model 
! surface fossil magnetic fields in massive star evolutionary models.
!
! Some of the new routines included below were part of the PhD Thesis of 
! Z. Keszthelyi 2019, Queen's Univesity, Kingston, Canada. 
!  
! The first application was presented by Petit, Keszthelyi, MacInnis, 
! et al. 2017, MNRAS, 540, 466, 1052 in the case of non-rotating models. 
! Since then, three published papers, using the Geneva code, 
! adopted a similar, coherent picture (Georgy, et al. 2017, A&A, 599, 5; 
! Song, et al. (2018) A&A, 609, 3; 
! and Keszthelyi, Meynet, Georgy, et al. 2019, MNRAS, 485, 5843 - Paper I). 
! 
! The implementation of magnetic braking adopted in this module is described in detial 
! by Keszthelyi, Meynet, Shultz, et al. 2020 MNRAS - Paper II. The main differences 
! (concerning primarily the SURF models) are described in this file and in Paper IV. 
! An important difference is that here (Paper IV), we use the equatorial magnetic field strength B_eq.
!
! Further models, including various magnetic field evolution scenarios were tested in the context
! of the magnetic B-tye star tau Sco (Keszthelyi, Meynet, Martins, et al. 2021, MNRAS - Paper III). 
! 
! The module is applicable to model known magnetic stars that possess surface fossil magnetic fields.
! Those fields form a large-scale magnetoshpere (see, e.g., Donati & Landstreet 2009, ARA&A, 47, 333, for a review). 
!
! Following the works of ud-Doula & Owocki 2002, ApJ, 576, 413;
! ud-Doula, Owocki, Townsend 2008, MNRAS, 385, 97;
! and ud-Doula, Owocki, Townsend 2009, MNRAS, 392, 1022,
! we adopt a formalism, which is applicable in one-dimensional codes.
!
! We consider the following ingredients to model surface fossil magnetic fields: 
!
! Mass-loss quenching reduces the effective mass-loss rate of the star (other_wind routine).
! Magnetic braking reduces the angular momentum of the star (other_torque routine).
! The fossil field evolves by time, following the frozen flux theorem (Alfven, 1942). 
!
! MAIN ADDITIONS in Paper IV (Keszthelyi, de Koter, Goetberg, et al. 2022, MNRAS) :
!
! Angular momentum transport via other_am_mixing: here, we adopt a simplified treatment. 
! The INT models are set in the inlist files, whereas the SURF models are set here in the 
! run_star_extras.
!
! Chemical mixing: here, we implement the Zahn et al. equations for the "Mix2" scheme. 
!
!
! We caution that there are several uncertainties and 
! this approach needs to be further improved.
! 
! Zsolt Keszthelyi et al. 2022
!
!==================================================================================
module run_star_extras

  use const_lib
  use const_def      
  use star_private_def  ! THIS ALWAYS HAS TO BE COPIED MANUALLY AFTER NEW INSTALLATION. It is in star/make and has to be in ../include
  use star_lib
  use star_def
  use chem_def, only: ih1, ihe4, ic12, ic13, in14, io16

  implicit none

      ! Module variables: 
      !==========================
      real(dp) :: R_init  ! initial radius, calc. actual value in mag routine
      real(dp) :: Beq     ! equatorial magnetic field strength 
      real(dp) :: B0      ! Assumed *equatorial* initial field (add in inlist, star_job) -- uncomment this if want to use initial flux
      real(dp) :: Beff    ! efficiency of magnetic braking - scaling factor
      real(dp) :: Deceff  ! efficiency of field decay - scaling factor
      real(dp) :: vesc    ! escape velocity (calculated for electron scattering opacity)
      real(dp) :: vinf    ! terminal velocity
      real(dp) :: etastar ! equatorial wind magnetic confinement parameter
      real(dp) :: Ra      ! Alfven radius
      real(dp) :: Rc      ! closure radius  
      real(dp) :: Rk      ! Kepler co-rotation radius
      real(dp) :: f_B     ! escaping mass fraction allowed by the magnetosphere
      real(dp) :: logf_B  ! log of escaping mass fraction
      real(dp) :: f_rot     ! rotational enhancement following Maeder & Meynet (2000)
      real(dp) :: Mdot_orig ! original mass-loss rate (unaffected by the magnetosphere) 
      real(dp) :: w_vink    ! original Vink+2001 mass-loss rates
      real(dp) :: w_vanloon ! original Van Loon+2005 mass-loss rates 
      real(dp) :: Jbrake    ! dJ / dt (total rate of angular momentum loss) from ud-Doula et al.  
      real(dp) :: J_surf    ! AM in the near-surface of the star 
      real(dp) :: scale     ! to (uniformly) distribute AM loss
      real(dp) :: core_h1_init  ! initial core H content
      real(dp) :: enclosedm  ! mass not affected by mag. torque
      integer  :: nin, k_max, ! these are used to determine the depth of magnetic braking
      integer, parameter :: UNIFORM_TORQUE = 1    ! the entire star slows its rotation
      integer, parameter :: SURFACE_TORQUE = 2    ! the surface layers slow their rotation
      integer, save  :: torque_method = 0 ! Method to evaluate torque
      integer, parameter :: CONST = 1    ! Magnetic field strength is constant
      integer, parameter :: CON = 2      ! Magnetic flux is conserved (propto R^-2)
      integer, parameter :: DEC = 3      ! Magnetic field strength decays
      integer, save  :: fieldevolution = 0 ! Field evolution scenario
      ! rotmix paramseters
      integer, parameter :: rasize = 19000 ! assume max number of zones
      real(dp) :: mod_U(rasize) ! radial dependence of the vertical component of meridional circulation
      real(dp) :: mod_D_shear(rasize)  ! shear mixing
    ! combination of meridional currents and strong horizontal turbulence => D_eff
      real(dp) :: mod_D_eff(rasize)      ! for chemical mixing
      real(dp) :: mod_D_h(rasize)            ! diffusion due to horizontal turbulance (denominator in D_eff)   
      real(dp) :: mod_dlnomega_dlnr(rasize)  ! shear param.
      real(dp) :: extra_jdot(rasize)
     ! for assumed magnetic field strength (radially)
      real(dp) :: Br(rasize)  ! 
      !==========================

 contains
      
   subroutine extras_controls(id, ierr)

      integer, intent(in) :: id
      integer, intent(out) :: ierr

      type (star_info), pointer :: s

      ierr = 0

      call star_ptr(id, s, ierr)
      if (ierr /= 0) return

      s% extras_startup => extras_startup
      s% extras_check_model => extras_check_model
      s% extras_finish_step => extras_finish_step
      s% extras_after_evolve => extras_after_evolve
      s% how_many_extra_history_columns => how_many_extra_history_columns
      s% data_for_extra_history_columns => data_for_extra_history_columns
      s% how_many_extra_profile_columns => how_many_extra_profile_columns
      s% data_for_extra_profile_columns => data_for_extra_profile_columns

      s% other_torque => MAG_braking          ! torque_routine
      s% other_wind => MAG_WIND               ! wind_routine
      s% other_am_mixing => MAG_MIX           ! simple version to enforce efficient AM transport
      s% other_d_mix => MAG_DMIX              ! chemical mixing following Zahn et al 1992

      ! Once you have set the function pointers you want,
      ! then uncomment this (or set it in your star_job inlist)
      ! to disable the printed warning message,

      s% job% warn_run_star_extras =.false.

      ! ====================================================
      ! Set magnetic parameters in the inlist file (star_job)
      ! B0 is the initial polar magnetic field strength in gauss
      ! The torque method will apply the torque for either the entire star (UNIFORM)
      ! or the surface AM reservoir (SURFACE)
      ! ====================================================

      if (s%job%extras_lrpar >= 1) then
         B0 = s%job%extras_rpar(1)   ! set as the EQUATORIAL FIELD STRENGTH
         Beff = s%job%extras_rpar(2)
         Deceff = s%job%extras_rpar(3)
      else
         stop 'Initial field strength and/or braking efficiency is not set using extras_rpar(1) and/or extras_rpar(2)'
      endif

      if (s%job%extras_lcpar >= 1) then

         !  if (s%job%extras_lcpar == 1) then
         select case (s%job%extras_cpar(1))
          case ('UNIFORM_TORQUE')
            torque_method = UNIFORM_TORQUE
          case ('SURFACE_TORQUE')
            torque_method = SURFACE_TORQUE
          case default
            stop 'Torque method is not set.'
         end select
         !   else
         !      stop 'Torque method is not set using extras_cpar(1)'
         !   endif
         !  if (s%job%extras_lcpar == 2) then

         select case (s%job%extras_cpar(2))
          case ('CONST')
            fieldevolution = CONST
          case ('CON')
            fieldevolution = CON
          case ('DEC')
            fieldevolution = DEC
          case default
            stop 'Field evolution is not set.'
         end select
         !   else
         !      stop 'Field evolution is not set using extras_cpar(1)'
         !  endif

      endif

   end subroutine extras_controls

   subroutine extras_startup(id, restart, ierr)
      integer, intent(in) :: id
      logical, intent(in) :: restart
      integer, intent(out) :: ierr
      integer :: h1
      type (star_info), pointer :: s
      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return

      ! Usual startup stuff
      if (.not. restart) then
         call alloc_extra_info(s)
      else ! it is a restart
         call unpack_extra_info(s)
      end if

      ! Store the initial H core mass fraction
      ! This will be used to calculate the ZAMS radius
      ! for magnetic flux conservation
      h1 = s% net_iso(ih1)
      core_h1_init = s% xa(h1,s% nz)

      R_init = 0.0d0 !!  R_init is set to zero for the first model.
      !! it is then calculated in the magnetic routine.

      WRITE (*,*) ':::--- Initial surface magnetic field strength in G ---:::'
      WRITE (*,*)  B0
      WRITE (*,*) '======================================================='
      WRITE (*,*) 'Torque method selected (uniform or surface)'
      WRITE (*,*)  torque_method
      WRITE (*,*) '======================================================='
      WRITE (*,*) 'Field evolution scenario (const. or flux conservation or field decay)'
      WRITE (*,*)  fieldevolution
      WRITE (*,*) '======================================================='

   end subroutine extras_startup

  ! ====================================================
  
   ! returns either keep_going, retry, or terminate.
   integer function extras_check_model(id)
      integer, intent(in) :: id
      integer :: ierr
      type (star_info), pointer :: s
      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return
      extras_check_model = keep_going
      !if (.false. .and. s% star_mass_h1 < 0.35d0) then
      ! stop when star hydrogen mass drops to specified level
      !   extras_check_model = terminate
      !   write(*, *) 'have reached desired hydrogen mass'
      !   return
      !end if  ! was this supposed to be default???
      ! by default, indicate where (in the code) MESA terminated
      if (extras_check_model == terminate) s% termination_code = t_extras_check_model
   end function extras_check_model

   integer function extras_finish_step(id)
      integer, intent(in) :: id
      integer :: ierr
      integer :: h1, k
      real(dp) :: core_h1
      type (star_info), pointer :: s
      real(dp), parameter :: F_LIMIT = 0.20_dp 

      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return

      extras_finish_step = keep_going
      call store_extra_info(s)
      call MAG_braking(id, ierr)

      ! ====================================================
      ! Time-step control implemented by R. Townsend
      ! Limit dt_next based on the torquing factor:
      ! This avoids possible issues with exhausting or exceeding angular momentum in a given layer
      ! and therefore reduces the time-step if necessary.
      ! This might cause the simulation to become very computational time-expensive!
      ! ====================================================
      if (scale > F_LIMIT) then
         s%dt_next = s%dt_next/(scale/F_LIMIT)
         WRITE(*,*) '====----------====='
         WRITE(*,*) 'A better next time step is:'
         WRITE(*,*) s%dt_next/secyer
         WRITE(*,*) '====----------====='
      endif

      h1 = s% net_iso(ih1)
      core_h1 = s% xa(h1,s% nz)

      !! ======= Profile for models at any custom X_c ==========
      if (core_h1 <= (core_h1_init * 0.985d0) .AND. core_h1 > (core_h1_init * 0.981d0)) then
         call star_write_profile_info(s% id, 'LOGS/prof_ZAMS.data', ierr)
         WRITE(*,*) 'A structure profile is saved.'
      end if
      if (core_h1 <= (core_h1_init * 0.550d0) .AND. core_h1 > (core_h1_init * 0.499d0)) then
         call star_write_profile_info(s% id, 'LOGS/prof_midMS.data', ierr)
         WRITE(*,*) 'A structure profile is saved.'
      end if
      if (core_h1 <= (core_h1_init * 0.09d0) .AND. core_h1 > (core_h1_init * 0.06d0)) then
         call star_write_profile_info(s% id, 'LOGS/prof_TAMS.data', ierr)
         WRITE(*,*) 'A structure profile is saved.'
      end if
      ! see extras_check_model for information about custom termination codes
      ! by default, indicate where (in the code) MESA terminated
      if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step

   end function extras_finish_step

!=========================================================
!  ******** Main additions from here *******************
!
! Note that since we need vinf, we calculate it explicitly from 
! the Eddington parameter for pure electron scattering.
! We also adopt the bi-stability, that is the behaviour of vinf/vesc. and Mdot at a given Teff.
! This should be carefully checked to be consistent with the wind description.

! === Terminal velocity using Gamma for electron scattering (Kudritzki et al. 1986)=== 
! === Rotational enhancement following Maeder & Meynet (2000) === 
! 
! gammae routine based on the implementation of 
! Keszthelyi, Puls, Wade, 2017, A&A 584, A4 (KPW2017)

   subroutine gammae(id,ierr)
      type (star_info), pointer :: s
      integer, intent(in) :: id
      integer, intent(out) :: ierr
      integer :: h1, he4, nz
      real(dp) :: alfa1, alfa2, L1, M1, T1, R1 ! alfa1,2 used for interpolation only
      real(dp) :: I_he, YHe, sigma_e, Gamma_e  ! Gamma_e is the Eddington factor for electron scattering
      real(dp) :: vinf1, vinf2, vinf3, logvinf
      real(dp) :: surface_h1, surface_he4
      real(dp) :: nom, denom, powe
      real(dp) :: f_rot1, f_rot2, f_rot3
      real(dp) :: X, Y, Z  ! H, He and metal mass fractions
      real(dp) :: alpha  ! force multiplier parameter with simple Teff dependence - new -
      ! --- Values related to the bi-stability --- check consistency with wind routine!

      ! The calculations are partitioned into three Teff regimes.
      real(dp), parameter :: Teff_jump1 = 20.d3   ! Revised bi-stability jump temperature. See Petrov, et al. (2016),
      ! KPW2017, and Vink (2018).
      ! FIRST bi-stability jump temperature (from observations)
      real(dp), parameter :: Teff_jump2 = 10.d3
      ! SECOND bi-stability jump temperature (from observations)
      real(dp), parameter :: dT1 = 1.d3   ! adopted for consistency with Vink routine
      ! Interpolation width of the first bi-stability jump.
      ! Note that a small dT results in a steep change
      ! in Mdot while a larger dT yields a more gradual change.
      real(dp), parameter :: dT2 = 1.d3   ! second jump
      ! Interpolation width of the second bi-stability jump.

      ! --- Values related to terminal wind velocity ---
      ! adopted from Kudritzki & Puls (2000), ARA&A, 38, 613; and Markova & Puls (2008), A&A, 478, 823
      real(dp), parameter :: fvinf1 = 2.6d0  ! ratio of terminal to escape velocity before Teff_jump_1
      real(dp), parameter :: fvinf2 = 1.3d0  ! ratio of terminal to escape velocity after Teff_jump_1 and before Teff_jump_2
      real(dp), parameter :: fvinf3 = 0.7d0  ! ratio of terminal to escape velocity after Teff_jump_2

      call get_star_ptr(id,s,ierr)

      L1 = s% photosphere_L * Lsun  ! cgs
      M1 = s% star_mass * Msun      ! cgs
      T1 = s% Teff                  ! cgs
      R1 = s% photosphere_r * Rsun  ! cgs

      !--- elements ----
      nz = s% nz
      h1 = s% net_iso(ih1)
      he4 = s% net_iso(ihe4)
      surface_h1 = s% xa(h1,1)
      surface_he4 = s% xa(he4,1)
      X = surface_h1            ! mass fraction of H
      Y = surface_he4           ! mass fraction of He
      Z = 1.d0 - (X + Y)        ! surface metallicity

      YHe = Y/X / 4.d0
      ! surface Helium content by number,
      ! n_He/n_H; approximation: factor 4 should be m_He/m_H

      ! === PATH 1 ===
      ! --- vinf behaviour --- check for consistency with wind routine
      if (T1 .ge. Teff_jump1 - dT1) then
         I_he = 2.d0  ! the number of free electrons per He nuclei (hot regime)
         sigma_e = 0.398d0 * (1.d0 + YHe * I_he) / (1.d0 + 4.d0 * YHe)  ! Eq. 66 by Kudritzki et al. (1989) A&A, 219, 205
         Gamma_e =  sigma_e * L1/(M1 * pi4 * standard_cgrav * clight)
         ! escape velocity corrected for electron scattering
         vesc = sqrt((1.d0 - Gamma_e) * 2.d0 * standard_cgrav * M1 / R1 )
         ! Eq. 15 KPW2017
         vinf1 = vesc * fvinf1
         vinf = vinf1
         ! rotational enhancement
         alpha = 0.6 ! set to 0.6 for path 1
         nom = 1.d0 - Gamma_e
         denom = 1.d0 - (4.d0/9.d0 * pow(s%v_rot_avg_surf / s%v_crit_avg_surf, 2.0d0) ) - Gamma_e  ! Note v_crit definition is different!
         powe = (1.d0 / alpha) - 1.d0
         f_rot1 = pow(nom, powe) / pow(denom, powe)
         f_rot = f_rot1
      endif
      ! === PATH 2 ===
      if (T1 .le. Teff_jump1 + dT1 .and. T1 .ge. Teff_jump2 - dT2) then
         I_he = 1.d0     ! cooler side of jump, Helium recombined to HeII
         sigma_e = 0.398d0 * (1.d0 + YHe * I_he) / (1.d0 + 4.d0 * YHe)
         Gamma_e =  sigma_e * L1/(M1 * pi4 * standard_cgrav * clight)
         vesc = sqrt(2.d0 * standard_cgrav * M1 / R1 * (1.d0-Gamma_e))
         vinf2 = vesc * fvinf2
         vinf = vinf2
         ! rotational enhancement
         alpha = 0.5 ! set for path 2
         nom = 1.d0 - Gamma_e
         denom = 1.d0 - (4.d0/9.d0 * pow(s%v_rot_avg_surf / s%v_crit_avg_surf, 2.0d0) ) - Gamma_e
         powe = (1.d0 / alpha) - 1.d0
         f_rot2 = pow(nom, powe) / pow(denom, powe)
         f_rot = f_rot2
      endif
      ! --- INTERPOLATION for the FIRST bi-stability jump ---
      if (Teff_jump1 + dT1 > T1 .and. T1 > Teff_jump1 - dT1) then
         alfa1 = (T1 - (Teff_jump1 - dT1)) / (2.d0 * dT1)
         vinf = alfa1 * vinf1 + (1.d0 - alfa1) * vinf2
         f_rot = alfa1 * f_rot1 + (1.d0 - alfa1) * f_rot2
      endif
      ! === PATH 3 ===
      if (T1 .le. Teff_jump2 + dT2) then
         sigma_e = 0.398d0 * 1.d0 / (1.d0 + 4.d0 * YHe) ! since I_he = 0 for T1 < 10 kK.
         Gamma_e =  sigma_e * L1/(M1 * pi4 * standard_cgrav * clight)
         vesc = sqrt(2.d0 * standard_cgrav * M1 / R1 * (1.d0-Gamma_e))
         vinf3 = vesc * fvinf3
         vinf = vinf3
         ! rotational enhancement
         alpha = 0.4
         nom = 1.d0 - Gamma_e
         denom = 1.d0 - (4.d0/9.d0 * pow(s%v_rot_avg_surf / s%v_crit_avg_surf, 2.0d0) ) - Gamma_e
         powe = (1.d0 / alpha) - 1.d0
         f_rot3 = pow(nom, powe) / pow(denom, powe)
         f_rot = f_rot3
      endif
      ! --- INTERPOLATION for the SECOND bi-stability jump ---
      if (Teff_jump2 + dT2 > T1 .and. T1 > Teff_jump2 - dT2) then
         alfa2 = (T1 - (Teff_jump2 - dT2)) / (2.d0 * dT2)
         vinf = alfa2 * vinf2 + (1.d0 - alfa2) * vinf3
         f_rot = alfa2 * f_rot2 + (1.d0 - alfa2) * f_rot3
      endif

      ! To avoid numerical problems, if f_rot is too small (or rotation is very slow),
      ! just ignore the rotational enhahncement on the mass-loss rates.

      if (f_rot .le. 1.d-4 .OR. s%v_rot_avg_surf .le. 1.d0) then
         f_rot = 1.d0
      endif

   end subroutine gammae

! ========================================================================
! === Calculation of MAGNETIC parameters including field evolution === 
!
! Initial routine used by:
! Petit, Keszthelyi, MacInnis, et al. 2017, MNRAS, 466, 1052,
! Keszthelyi, Meynet, Shultz, et al. 2020, MNRAS, Volume 493, 1, 518-535
! and Keszthelyi, Meynet, Martins, de Koter, David-Uraz, 2021, MNRAS, Volume 504, 2,2474-2492
!
! Updates are described in Paper IV (Keszthelyi, de Koter, Goetberg, 2022, MNRAS). 
! ========================================================================

  subroutine magnetic(id,ierr)
      type (star_info), pointer :: s
      integer, intent(in) :: id
      integer, intent(out) :: ierr
      integer :: h1
      real(dp) :: L1, M1, T1, R1, core_h1
      real(dp) :: logB0, logBp, logBeq, logetastar
      real(dp) :: logvesc, logvinf
      integer :: k

      !====  Initial magnetic flux can also be set.
      !      Set only if B0 is not set.
      !  real(dp), parameter :: F0 = 10**28    ! initial magnetic flux in units of G cm2
      !====

      call gammae(id,ierr)  ! call in gammae for vinf routine
      call get_star_ptr(id,s,ierr)

      ! basic parameters
      L1 = s% photosphere_L * Lsun  ! cgs
      M1 = s% star_mass * Msun      ! cgs
      T1 = s% Teff                  ! cgs
      R1 = s% photosphere_r * Rsun  ! cgs
      h1 = s% net_iso(ih1)
      core_h1 = s% xa(h1,s% nz)
      ! ====================================
      ! Setting the ZAMS radius here.
      ! We adopt this ZAMS definition: some hydrogen is already fused in the core
      ! to generate the stellar luminosity.
      ! Initiate magnetic flux conservsation
      ! set radius when rotation relaxed -> then set B0 to that value too.
      ! value for relaxation is chosen to be 50 in inlists.
      if (core_h1 <= (core_h1_init * 0.997d0) .AND. R_init <= 0.d0) then
         R_init = s% photosphere_r * 6.9598d10
      end if
      ! ====================================
      if (.not. s% doing_relax) then       ! don't do calc. for relaxation phase

      ! If the initial flux F0 is set instead of B0, then B0 needs to be calculated.
      ! F0 and R_init are in cgs, pi4 = 4. * pi
      ! B0 = F0/(pi4 * R_init * R_init)    ! this equation considers the unsigned flux.

         select case (fieldevolution)  ! set in inlist.
         ! =====================================
         !  ===  MAGNETIC FIELD CONSTANT ===
         ! =====================================
          case (CONST)
            Beq = B0
            logBeq = log10(B0)
         ! =====================================
         !  ===  MAGNETIC FLUX CONSERVATION ===
         ! (considered for equatorial field strength)
         ! =====================================
          case (CON)
            logBeq = log10(B0) + 2.d0 * log10(R_init) - 2.d0 * log10(R1)  ! cgs
            Beq = exp10(logBeq)
         ! =====================================
         !  ===  MAGNETIC FLUX DECAY ===
         ! =====================================
          case (DEC) ! avoid for now. MS age needs to be set up for different masses.
            !Bp = B0 * exp10_cr(Deceff*s%star_age/12d6)
            Beq = B0*exp10(-Deceff*s%star_age/12d6)  ! 12Myr was tailored for tau Sco's assumed MS age in Paper III.
            logBeq = log10(Beq)                      ! star_age is in units of yr.

          case default
            stop 'Field is not evolving.'
         end select  ! method

         ! ====================================
         ! === Equatorial magnetic confinement parameter ===   ! in cgs
         logetastar = 2.d0 * logBeq + 2.d0 * log10(R1)    &
            - log10(Mdot_orig * Msun/secyer)   &
            - log10(vinf)
         etastar = exp10(logetastar)

         ! === Alfven radius ===
         ! depending on assumed field geometry - dipole for INT, quadrupole for SURF
         select case (torque_method)  ! set in inlist.
          case (UNIFORM_TORQUE) ! 1/4 dependence for dip.
            Ra = 1.0d0 + pow(etastar + 0.25d0, 0.25d0) - pow(0.25d0, 0.25d0)      ! this is: R_A / R_star
          case (SURFACE_TORQUE) ! 1/6 dependence for quad.
            Ra = 1.0d0 + pow(etastar + 0.25d0, 0.1667d0) - pow(0.25d0, 0.1667d0)  ! this is: R_A / R_star
          case default
            stop 'Invalid torque method'
         end select  ! method

         !==== closure and Kepler co-rotation radii ===
         Rc = 1.0d0 + 0.7d0 * (Ra - 1.0d0)                                      ! this is: R_c / R_star
         Rk = pow(s% v_rot_avg_surf / sqrt(standard_cgrav*M1/R1), -2.d0/3.d0)   ! this isL R_K / R_star

         ! === Mass-loss quenching parameter ===    ! used to multiply original Mdot in wind routine
         if (Ra < Rk) then  ! definition of a dynamical magnetoshpere (DM)
            f_B = 1.0d0 - sqrt(1.0d0 - 1.0d0/Rc)
            logf_B = log10(f_B)
         else  ! Rk .leq. Ra ! definition of a centrifugal magnetosphere (CM)
            f_B = 2.0d0 - sqrt(1.0d0 - 1.0d0/Rc) - sqrt(1.0d0 - 0.5d0/Rk)
            logf_B = log10(f_B)
         end if
         ! For rotating models, this scaling considers
         ! an updated version (compared to Paper II), using Eq. 22 from ud-Doula et al. (2009).
         ! DMs will use the usual scaling.


         if (s%r(1) > 0) then
         ! === assume internal field strengths ===
         ! depending on assumed field geometry - dip for INT, quad for SURF
            select case (torque_method)  ! set in inlist.
             case (UNIFORM_TORQUE) ! 1/3 radial dependence for dip.
               do k=1,s%nz
                  Br(k)= Beq*2.d0 * pow(s%r(1)/s%r(k),3.d0)  ! factor of 2 for eq to polar
               end do
             case (SURFACE_TORQUE) ! 1/4 radial dependence for quad.
               do k=1,s%nz
                  Br(k)= Beq*4.d0 * pow(s%r(1)/s%r(k),4.d0)  ! factor of 4 for eq to quad
               end do
             case default
               stop 'Invalid torque method'
            end select  ! method
         end if

      endif  ! for not doing relax.

   end subroutine magnetic

! ====================================
!      === MAGNETIC BRAKING === 
! ====================================

   subroutine MAG_braking(id, ierr)
      integer, intent(in) :: id
      real(dp) :: R1, M1, T1, Mdot, logvesc, I_surf
      real(dp) :: logvinf, logetastar, Omega
      integer, intent(out)      :: ierr
      type (star_info), pointer :: s
      integer                   :: k, nz
      real(dp)                  :: J_lost, J_ex
      ierr = 0
      call get_star_ptr(id, s, ierr)
      if (ierr /= 0) return

      if (s% doing_relax) then
         write(*,*) 'still relaxing a bit - you should do the same'
         write(*,*) s% star_age
      end if

      if (.not. s% doing_relax .AND. R_init > 0._dp .AND. s%mstar_dot < 0._dp .AND. s%omega(1) /= 0._dp) then

         R1 = s% photosphere_r * Rsun  ! cgs
         M1 = s% photosphere_m * Msun  ! cgs

         do k = 1, s%nz ! find where BRAKING and TRANSPORT is applied from
            if (s% q(k) .LE. 0.80d0) then !
               nin = k  ! index of the last zone where braking is applied
               exit
            endif
         end do
         k_max = nin

         if (nin > 0.) then
            enclosedm = sum(s%dm_bar(1:k_max)) / Msun  ! this is the mass of the layers where braking is applied
         else
            enclosedm = 0.d0
         end if

         Mdot =  Mdot_orig * Msun / secyer        ! in cgs - this is the original mass-loss rate for B=0.
         Omega = s% omega(1)                      ! or s% omega_avg_surf  ! in history this is 'surf_avg_omega'
         J_lost = s%angular_momentum_removed      ! by the mass loss

         ! === Total angular momentum loss due to magnetic fields ===
         ! This is dJ(field+gas)/dt from ud Doula, Owocki, Townsend (2009)
         ! Ra is in R_star units, so now it is converted to cgs.
         ! The negative sign will follow later to extract this quantity.

         Jbrake = (2.0d0 / 3.0d0) * Mdot * Omega * (Ra*R1)**2.d0    ! this is dJ/dt
         J_ex = max(Jbrake * s% dt - J_lost,0._dp)   ! avoid double counting the AM removed by mass-loss.

         if (J_ex < 0._dp) then
            WRITE (*,*) ':::::::::::::'
            WRITE (*,*) 'J ex cannot be negative.'
            WRITE (*,*) Mdot, Omega, s% v_rot_avg_surf, Ra, Jbrake, J_ex
            ierr = -1
            return
         end if

         if (J_ex >= 0._dp) then
         ! ==================================================================
         ! Two torque methods are introduced because the exact
         ! penetration depth of the fossil field remains unknown.
         !
         ! UNIFORM will remove AM from the entire star, SURFACE will only brake the rotation in some layers.
         ! This is a simplifying assumption, of course.

            select case (torque_method)   ! set in inlist.
             
             case (UNIFORM_TORQUE)        ! convert J_ex to extra_j throughout the entire star.
               ! Jbrake converted to specific AM via scale:
               ! the same fraction of specific AM is removed from layer to layer (but not the same amount).
               scale = Beff * J_ex / s% total_angular_momentum ! consider efficiency parameter Beff (in inlist)

               if (scale < 1._dp .AND. scale > 1.d-15) then
                  do k = 1,s%nz-2  ! from surface to all layers down to the stellar core
                     s%extra_jdot(k) = - s% j_rot(k) *  scale / s% dt    ! negative sign added here: this is AM loss.
                     extra_jdot(k) = s%extra_jdot(k)
                     !===  inlist controls are used to enforce co-rotation in addition in this case
                  end do
               else
                  do k = 1, s%nz
                     s% extra_jdot(k) = 0._dp
                     extra_jdot(k) = s%extra_jdot(k)
                  end do
               end if

             case (SURFACE_TORQUE)        ! convert J_ex to extra_j in the near-surface of the star.

               J_surf = DOT_PRODUCT(s%dm_bar(1:k_max), s%j_rot(1:k_max))
               I_surf = DOT_PRODUCT(s%dm_bar(1:k_max), s%i_rot(1:k_max))

               scale = Beff * J_ex  / J_surf ! consider efficiency parameter Beff (in inlist)
               ! apply magnetic torque if the scaling is appropriate
               if (scale < 1._dp .AND. scale > 1.d-15) then
                  do k = 1,k_max
                     s%extra_jdot(k) = - s% j_rot(k) * scale / s% dt    ! negative sign added here: this is AM loss.
                     extra_jdot(k) = s%extra_jdot(k)   ! amount of magnetic braking per layer
                  end do
               else
                  do k = 1,k_max
                     s% extra_jdot(k) = 0._dp
                     extra_jdot(k) = s%extra_jdot(k)
                  end do
               end if

             case default
               stop 'Invalid torque method'
            end select  ! method

         end if      ! Jbrake not zero
      end if         ! not doing relaxation

   end subroutine MAG_braking

   !====================================================
   ! Angular momentum transport - for SURF models only!
   ! ===================================================
   ! AM transport for the INT models is set in the inlist. 
   ! other_am_nu is set .true. in the inlist only for SURF models. 
   ! this will lead to a flat Omega only in those layers where braking is also applied.
   subroutine MAG_MIX(id, ierr)
      integer, intent(in) :: id
      integer, intent(out) :: ierr
      type (star_info), pointer ::s
      integer :: k
      ierr = 0
      call get_star_ptr(id, s, ierr)
      if (ierr /= 0) return

      if (.not. s% doing_relax .AND. R_init> 0._dp .AND. s%mstar_dot < 0._dp .AND. s%omega(1) /= 0._dp) then
         do k=1,k_max  ! until k_max layer for SURF models (80% of mass fraction)
            s%am_nu_omega(k) = 1.d16   ! 
         end do
      end if

   end subroutine MAG_MIX

   subroutine magic(s, log_abs_shear)  ! this is the mesa version to calculate domega/dlnr 
      type (star_info), pointer :: s
      real(dp) :: log_abs_shear(:)
      real(dp) :: y1,y2,x1,x2,yyaa,yybb
      integer :: i_omega, i_lnR, k, nz, j
      real(dp), pointer :: r(:)
      real(dp) :: domega_dr

      ! ln radius at outer face of cell
      i_lnR = s% i_lnR
      nz = s% nz
      allocate(r(nz))
      ! allocate(mod_dlnomega_dlnr(nz))
      ! vectors of structure variables: xh(i,k) holds hydro structure variable i for point k
      do k=1,nz
         r(k) = exp(s% xh(i_lnR,k))
      end do

      do k=2,nz-1
         domega_dr = 2.d0*(s% omega(k-1) - s% omega(k))/(r(k-1) - r(k+1))
         mod_dlnomega_dlnr(k) = domega_dr*(r(k-1) + r(k+1))/(s% omega(k-1) + s% omega(k))
         ! log_abs_shear(k) = safe_log10_cr(max(1d-30,min(1d30,abs(mod_dlnomega_dlnr))))
      end do

      mod_dlnomega_dlnr(1) = mod_dlnomega_dlnr(2)
      mod_dlnomega_dlnr(nz) = mod_dlnomega_dlnr(nz-1)
      ! log_abs_shear(1) = log_abs_shear(2)
      ! log_abs_shear(nz) = log_abs_shear(nz-1)

      deallocate(r)
      ! deallocate(mod_dlnomega_dlnr)

   end subroutine magic

   ! =================================================================
   ! Chemical element transport following the work of Zahn 1992
   ! =================================================================
   ! implementation credit to Martin Carrington and Zsolt Keszthelyi

   subroutine MAG_DMIX(id, ierr) 
      implicit none
      integer, intent(in) :: id
      integer, intent(out) :: ierr
      integer :: k ! used to loop over zones of star
      real(dp) :: to_be_squared, KAY_numerator, KAY_denominator, KAY, square, square_brackets, delta
      real(dp) :: u1, u2, u3, U, m_star, alpha, eps_m, E_mu, g_tilde_over_g_bar, E_omega_star, D_h, V, rho_m
      real(dp) :: E_omega_star_term1, E_omega_star_term2, E_omega_star_term3 
      real(dp) :: Grav
      real(dp), pointer :: log_abs_shear(:)
      type (star_info), pointer :: s
      ! good approximation
      real(dp), parameter :: phi = 1.d0  ! dln(rho) / dln(mu) _ const P, T

      ierr = 0
      call get_star_ptr(id, s, ierr)
      if (ierr /= 0) return
      if (s% model_number < 1) return

      Grav = standard_cgrav

      call magic(s, log_abs_shear)

      if (.not. s% doing_relax .AND. s%omega(1) /= 0) then

         do k= 1,s% nz

            delta = s% chiT(k) / s% chirho(k) ! approx for: - (partial ln T / partial ln rho )_mu, P

            ! D_shear ==============================================================================================================================
            ! crad is "a", radiation constant
            KAY_numerator = 4.d0 * crad * clight * s% T(k)**4. * s% grada(k)   ! grada is adiabatic gradient
            KAY_denominator = 3.d0 * s% opacity(k) * s% rho(k) * s% P(k) * delta
            KAY = KAY_numerator / KAY_denominator

            to_be_squared = 9.d0*pi/32.d0 * s% omega(k) * mod_dlnomega_dlnr(k)
            square = to_be_squared * to_be_squared

            square_brackets = s% gradL_composition_term(k) + s% grada(k) - s% gradr(k)  ! gradr is nabla_rad

            ! In star_utils.f90: s% grav(k) = s% cgrav(k)*s% m_grav(k)/(s% r(k)*s% r(k)) where cgrav is gravitational constant and m_grav the enclosed mass at cell outer edge
            ! scale_heigt is in cm. Defined in star_utils, Hp = P_face/(rho_face*s% grav(k)),    s% scale_height(k) = min(Hp, alt_Hp)

            mod_D_shear(k) = (s% scale_height(k) /( s% grav(k) * delta) )* (KAY / square_brackets) * square

            call check_real(mod_D_shear(k),'D_shear')

            ! D_eff =========================================================================================================================
            ! U(r) is the radial dependence of the vertical component of the meridional circulation velocity
            ! Eqn 10 by Song et al 2013, eqn 2 by Ekstroem et al 2012
            ! divide equation into 3 factors, u1,u2,u3 for clarity
            u1 = s% P(k) / s% rho(k) / s% grav(k) / s% Cp(k) / s% T(k)
            u2 = 1.d0 / (s% grada(k) - s% gradr(k) + s% gradL_composition_term(k))
            ! M_star = s% m(k) * (1d0 - ( s% omega(k) ** 2 / (2d0 * pi * s% grav(k) * s% rho(k) / s% m(k)) ) )  ! rho_m in the equation, not rho, assuming internal average,
            rho_m = s% m(k) / (4.d0/3.d0 * pi * s% r(k) ** 3.d0)
            M_star = s% m(k) * (1.d0 - s% omega(k) **2 / 2.d0 / pi / s% cgrav(k) / rho_m)
            ! like eps_m, see below eqn 4.25 Maeder and Zahn 1998
            eps_m = s% L(k) / s% m(k) ! negligible according to Maeder and Zahn 1998
            g_tilde_over_g_bar = 4.d0 / 3.d0  * s% omega(k) **2 * s% r(k)**3 / Grav / s% m(k)   ! eqn 4.26 by Maeder and Zahn 1998
            ! THETA = 1d0 / 3d0 * s% r(k) ** 2 / s% grav(k) * 2d0 * s% omega(k) * (mod_dlnomega_dlnr(k) * s% omega(k) / s% r(k)) ! eqn. 4.27 Maeder and Zahn 1998
            ! E_omega_star = 2d0 * (1d0 - (s% omega(k) ** 2 / 2d0 / pi / s% cgrav(k) / s% rho(k))) * g_tilde_over_g_bar
            E_omega_star_term1 =  2.d0 * g_tilde_over_g_bar * (1.d0 - (s% omega(k) ** 2.d0 / 2.d0 / pi / Grav / s% rho(k)) )  ! &
            ! - (s% eps_nuc(k) + s% eps_grav(k)) / eps_m )   ! neglected (see above)
            E_omega_star_term2 = 0.d0  ! - rho_m / s% rho(k) * s% r(k) / 3.d0 .... ! set to zero for simplicity - see Maeder & Zahn 1998
            E_omega_star_term3 = 0.d0  ! set to zero - see Maeder and Zahn 1998
            E_omega_star = E_omega_star_term1 + E_omega_star_term2 + E_omega_star_term3
            E_mu = 0.d0  ! set to zero
            u3 = s% L(k) / M_star * (E_omega_star + E_mu)    ! neglecting last term which is small (Cp/delta * partial Theta / partial t)

            U = u1*u2*u3           ! radial component of meridional circulation
            mod_U(k) = U           ! store in module variable
            call check_real(U,'U')

            ! V(r) --------------------------------------
            V = U  ! diff.approx.    ! horizontal component of the meridional circulation velocity

            ! D_h ------------------------------------
            alpha = 1.d0 + 0.5d0 * mod_dlnomega_dlnr(k)
            D_h = s% r(k) * abs(2.d0 * V - alpha * U)
            mod_D_h(k) = D_h   ! store in module variable

            if (D_h == 0) then
               mod_D_eff(k) = abs(s% r(k) * mod_U(k))** 2  ! if division by zero
            else
               mod_D_eff(k) = 1.d0/30.d0 * abs(s% r(k) * mod_U(k))** 2 / mod_D_h(k) ! 1/30 is NOT to be confused with fc.
            end if
            ! Here, the numerical factor arises from integrating the Legandre polinomials -
            ! has nothing to do with calibrating scaling factors (see Chaboyer & Zahn 1992).
            call check_real(mod_D_eff(k),'D_eff')

            if (s% mixing_type(k) /= convective_mixing .and. &
               s% grada(k) + s% gradL_composition_term(k) - s% gradr(k) > 0.0 .and. &
               s% grada(k) - s% gradr(k) > 0.0) then  ! condition by Etienne Kaiser

               s% D_mix(k) = mod_D_eff(k) + mod_D_shear(k) ! final diffusion coefficient for chemial mixing in non-convective zones
            end if  ! for mixing type
         end do ! k=1,s% nz
      end if ! for no relax.

   end subroutine MAG_DMIX

   ! Martin Carrington
   subroutine check_real(var, var_name)
      real(dp) :: var
      character(*) :: var_name
      if (isnan(var) .or. (.not. var<0 .and. .not. var>=0) .or. var == 1d0/0 .or. var == -1d0/0) then ! isnan works with gnu compiler.
         print*,'check_real failed becaues of: ', var_name, var
         stop
      end if
   end subroutine check_real

   ! Etienne Kaiser
   subroutine check_isPositive(var, var_name)
      real(dp) :: var
      character(*) :: var_name
      if(var < 0.0) then
         print*,'check_isPositive failed: ', var_name, var
         stop
      end if
   end subroutine check_isPositive


   !=============================================================
   ! === 'Default' WIND routines: Vink+ and van Loon+ ===
   !=============================================================

   !--- MESA's default Vink routine,
   ! updated following Keszthelyi, Puls, Wade, 2017, A&A. --- 

   subroutine Vink(id, Lsurf, Msurf, Rsurf, Tsurf, X, Y, Z, w, ierr)
      use const_def, only: dp
      integer, intent(in) :: id
      real(dp), intent(in) :: Lsurf, Msurf, Rsurf, Tsurf, X, Y, Z ! surface values (cgs)
      integer, intent(out) :: ierr
      type (star_info), pointer :: s
      integer :: h1, he4, nz
      real(dp) :: L1, M1, R1, T1, surface_h1, surface_he4
      logical, parameter :: dbg = .false.
      real(dp), intent(out) :: w
      real(dp) :: alfa, w1a, w2a, w1, w2, w3, Teff_jump, logMdot, dT, vinf_div_vesc, logMdot_orig
      real(dp), parameter :: Zsolar = 0.019d0   ! Vink et al. use Z = 0.019 as solar

      call get_star_ptr(id,s,ierr)

      nz = s% nz
      h1 = s% net_iso(ih1)
      he4 = s% net_iso(ihe4)
      surface_h1 = s% xa(h1,1)
      surface_he4 = s% xa(he4,1)
      L1 = Lsurf   ! cgs
      M1 = Msurf   ! cgs
      T1 = Tsurf   ! cgs
      R1 = Rsurf   ! cgs

      ! needs to be consistent with vinf calc. in gammae subrutine !
      ! alfa = 1 for hot side, = 0 for cool side
      if (T1 > 21.d3) then   ! Note the updated bi-stability jump temperature !
         alfa = 1
      else if (T1 < 19.d3) then
         alfa = 0
      else ! previously used the Vink et al 2001, eqns 14 and 15 to set "jump" temperature
         ! 1d3*(61.2d0 + 2.59d0*(-13.636d0 + 0.889d0*log10_cr(Z/Zsolar)))
         Teff_jump = 20d3   ! Updated value, see KPW2017
         dT = 1.d3          ! changed, see KPW2017
         if (T1 > Teff_jump + dT) then
            alfa = 1
         else if (T1 < Teff_jump - dT) then
            alfa = 0
         else
            alfa = (T1 - (Teff_jump - dT)) / (2*dT)
         end if
      end if

      if (alfa > 0) then ! eval hot side wind (eqn 24)
         vinf_div_vesc = 2.6d0 ! this is the hot side galactic value
         vinf_div_vesc = vinf_div_vesc*pow(Z/Zsolar,0.13d0) ! corrected for Z
         logMdot = &
            - 6.697d0 &
            + 2.194d0*log10(L1/Lsun/1d5) &
            - 1.313d0*log10(M1/Msun/30) &
            - 1.226d0*log10(vinf_div_vesc/2d0) &
            + 0.933d0*log10(T1/4d4) &
            - 10.92d0*pow2(log10(T1/4d4)) &
            + 0.85d0*log10(Z/Zsolar)
         w1a = exp10(logMdot)               ! in Msun/yr
      else
         w1a=0
      end if

      if (alfa < 1) then ! eval cool side wind (eqn 25)
         vinf_div_vesc = 1.3d0 ! this is the cool side galactic value
         vinf_div_vesc = vinf_div_vesc*pow(Z/Zsolar,0.13d0) ! corrected for Z
         logMdot = &
            - 6.688d0 &
            + 2.210d0*log10(L1/Lsun/1d5) &
            - 1.339d0*log10(M1/Msun/30) &
            - 1.601d0*log10(vinf_div_vesc/2d0) &
            + 1.07d0*log10(T1/2d4) &
            + 0.85d0*log10(Z/Zsolar)
         w2a = exp10(logMdot)             ! in Msun/yr
      else
         w2a=0
      end if
      w = alfa*w1a + (1 - alfa)*w2a
      w_vink = 0.5 * w  !!! 0.5 introduced here as Vink rates *might* be overpredicting Mdot, according to several studies.

   end subroutine Vink

! ========================================
!   === MAGNETIC MASS-LOSS QUENCHING ===
!      === ROTATIONAL ENHANCEMENT === 
! ========================================

   subroutine MAG_WIND(id, Lsurf, Msurf, Rsurf, Tsurf, X, Y, Z, w, ierr)
      use const_def, only: dp
      integer, intent(in) :: id
      real(dp), intent(in) :: Lsurf, Msurf, Rsurf, Tsurf, X, Y, Z ! surface values (cgs)
      real(dp), intent(out) :: w ! wind in units of Msun/year (value is >= 0)
      integer, intent(out) :: ierr
      type (star_info), pointer :: s
      real(dp) :: T1, L1, M1, R1
      real(dp) :: w1, logMdot, gamma_edd, xsurf, beta, gammazero, lgZ ! for Gr. wind
      real(dp), parameter :: Zsolar = 0.019d0   ! for G. wind
      logical, parameter :: dbg = .false.
      real(dp) :: log10wvan
      include 'formats'

      call get_star_ptr(id,s,ierr)  ! pointer to get data
      call magnetic(id,ierr)  ! quenching routine to obtain f_B
      call gammae(id,ierr)    ! call for rotational boost on mass-loss rates 

      T1 = Tsurf
      L1 = Lsurf
      M1 = Msurf
      R1 = Rsurf

      if (T1 .GE. 10.d3 .AND. X > 0.4d0) then 
         call Vink(id, Lsurf, Msurf, Rsurf, Tsurf, X, Y, Z, w, ierr)
         Mdot_orig = w_vink
      end if
 
      ! helium rich Wolf-Rayet star: Nugis & Lamers
      if (T1 .GE. 10.d3  .AND. X < 0.4d0) then 
         Mdot_orig =  1.d-11 * pow(L1/Lsun,1.29d0) * pow(Y,1.7d0) * sqrt(Z)
      end if

! cool wind 
      if (T1 < 10.d3) then  
         log10wvan = -5.65d0 + 1.05*log10(L1/(1d4*Lsun)) - 6.3d0*log10(T1/35d2)
         w_vanloon = exp10(log10wvan)
         Mdot_orig = w_vanloon
      end if 

! de Jager, C., Nieuwenhuijzen, H., & van der Hucht, K. A. 1988, A&AS, 72, 259.
! if (X > 0.4d0 .and. T1 < 10.d3) then  
!  logMdot = 1.769d0*log10_cr(L1/Lsun) - 1.676d0*log10_cr(T1) - 8.158d0
!  w = exp10_cr(logMdot) 
!  w_dejager = w 
!  Mdot_orig = w_dejager
! end if  

 ! ========================================
 ! === Final mass-loss rate of the star ===
 ! ========================================
      if (f_B > 0.d0 .AND. f_rot > 0.d0) then 
         w = Mdot_orig * f_B * f_rot  ! scaled by mass-loss quenching and rotational enhancement
      else 
         w = Mdot_orig 
         write(*,*) 'Either f_B or f_rot is zero.' 
         write(*,*) f_B, f_rot
      end if
      if (dbg) write(*,*) 'wind', w

   end subroutine MAG_WIND

! =====================

   integer function how_many_extra_history_columns(id)
      integer, intent(in) :: id
      integer :: ierr
      type (star_info), pointer :: s
      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return
      how_many_extra_history_columns = 15
   end function how_many_extra_history_columns

   integer function how_many_extra_profile_columns(id)
      integer, intent(in) :: id
      integer :: ierr
      type (star_info), pointer :: s
      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return
      how_many_extra_profile_columns = 7
   end function how_many_extra_profile_columns

   subroutine data_for_extra_history_columns(id, n, names, vals, ierr)
      integer, intent(in) :: id, n
      character (len=maxlen_history_column_name) :: names(n)
      real(dp) :: vals(n)
      type(star_info), pointer :: s
      integer, intent(out) :: ierr
      ierr = 0
      call star_ptr(id,s,ierr)
      if(ierr/=0) return
      ! note: do NOT add these names to history_columns.list
      ! the history_columns.list is only for the built-in log column options.
      ! it must not include the new column names you are adding here.

      ! extra column 1
      names(1) = "Beq"   ! equatorial magnetic field strength in gauss
      vals(1) = Beq

      ! extra column 2
      names(2) = "vesc" ! escape velocity in cgs
      vals(2) = vesc

      ! extra column 3
      names(3) = "vinf" ! terminal wind velocity in cgs
      vals(3) = vinf

      ! extra column 4
      names(4) = "etastar" ! equatorial magnetic confinement parameter
      vals(4) = etastar

      ! extra column 5
      names(5) = "Ra"     ! Alfven radius / stellar radius 
      vals(5) = Ra

      ! extra column 6
      names(6) = "Rc"     ! closure radius / stellar radius 
      vals(6) = Rc

      ! extra column 7
      names(7) = "f_B"    ! escaping wind fraction (quenching parameter)
      vals(7) = f_B

      ! extra column 8
      names(8) = "f_rot"  ! rotational enhancement on the mass-loss rates
      vals(8) = f_rot

      ! extra column 9    ! mass of the layers undergoing braking in the SURF models
      names(9) = "enclosedm"
      vals(9) = enclosedm

      ! extra column 10      ! index of last layer where braking is applied
      names(10) = "nin"
      vals(10) = nin

      ! extra column 11
      names(11) = "Jbrake"  ! dJ / dt from ud-Doula, Owocki, Townsend 2009
      vals(11) = Jbrake

      ! extra column 12
      names(12) = "Mdot_orig"  ! mass-loss rates before applying f_B and f_rot
      vals(12) = Mdot_orig

      ! extra column 13      
      names(13) = "w_vink"    ! nominal Vink et al. rates divided by a factor of 2
      vals(13) = w_vink

      ! extra column 14
      names(14) = "Rk"        ! Kepler co-rotation radius / R_star
      vals(14) = Rk

      ! extra column 15
      names(15) = "J_surf"     ! surface angular momentum reservior (20% of mass) in cgs
      vals(15) = J_surf

! If you add / remove columns, it has to be synchronized with history column number 

   ierr = 0
   end subroutine data_for_extra_history_columns

   subroutine data_for_extra_profile_columns(id, n, nz, names, vals, ierr)
      integer, intent(in) :: id, n, nz
      character (len=maxlen_profile_column_name) :: names(n)
      real(dp) :: vals(nz,n)
      integer, intent(out) :: ierr
      type (star_info), pointer :: s
      integer :: k
      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return
      call MAG_braking(id, ierr)
      call MAG_DMIX(id, ierr)

      !note: do NOT add the extra names to profile_columns.list
      ! the profile_columns.list is only for the built-in profile column options.
      ! it must not include the new column names you are adding here.
      ! here is an example for adding a profile column
      !if (n /= 1) stop 'data_for_extra_profile_columns'

      names(1) = 'extra_jdot'
      do k = 1, nz
         vals(k,1) = extra_jdot(k)  ! angular momentum removed in a given layer.
      end do
      
      names(2) = 'mod_D_shear'
      do k = 1, nz
         vals(k,2) = mod_D_shear(k)  ! diffusion coefficient for shear mixing - Zahn et al. 
      end do

      names(3) = 'mod_dlnomega_dlnr'
      do k = 1, nz
         vals(k,3) = mod_dlnomega_dlnr(k)  ! shear term
      end do

      names(4) = 'mod_U'
      do k = 1, nz
         vals(k,4) = mod_U(k)  ! meridional circulation - Zahn et al. 
      end do

      names(5) = 'mod_D_eff'
      do k = 1, nz
         vals(k,5) = mod_D_eff(k)  ! effective diffusion coeff. - Zahn et al. 
      end do

      names(6) = 'Br'
      do k = 1, nz
         vals(k,6) = Br(k)  ! assumed radial field strength per layer 
      end do

      names(7) = 'am_nu_omega'
      do k = 1, nz
         vals(k,7) =s% am_nu_omega(k) ! diffusivity for AM transport
      end do

   end subroutine data_for_extra_profile_columns
      
   subroutine extras_after_evolve(id, ierr)
      integer, intent(in) :: id
      integer, intent(out) :: ierr
      type (star_info), pointer :: s
      ierr = 0
      call star_ptr(id, s, ierr)
      if (ierr /= 0) return
   end subroutine extras_after_evolve

   subroutine alloc_extra_info(s)
      integer, parameter :: extra_info_alloc = 1
      type (star_info), pointer :: s
      call move_extra_info(s,extra_info_alloc)
   end subroutine alloc_extra_info
      
   subroutine unpack_extra_info(s)
      integer, parameter :: extra_info_get = 2
      type (star_info), pointer :: s
      call move_extra_info(s,extra_info_get)
   end subroutine unpack_extra_info

   subroutine store_extra_info(s)
      integer, parameter :: extra_info_put = 3
      type (star_info), pointer :: s
      call move_extra_info(s,extra_info_put)
   end subroutine store_extra_info

   subroutine move_extra_info(s,op)
      integer, parameter :: extra_info_alloc = 1
      integer, parameter :: extra_info_get = 2
      integer, parameter :: extra_info_put = 3
      type (star_info), pointer :: s
      integer, intent(in) :: op
    
      integer :: i, j, num_ints, num_dbls, ierr
         
      i = 0
      ! call move_int or move_flg    
      num_ints = i
         
      i = 0
      ! call move_dbl       
    
      num_dbls = i

      if (op /= extra_info_alloc) return
      if (num_ints == 0 .and. num_dbls == 0) return

      ierr = 0
      call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
      if (ierr /= 0) then
         write(*,*) 'failed in star_alloc_extras'
         write(*,*) 'alloc_extras num_ints', num_ints
         write(*,*) 'alloc_extras num_dbls', num_dbls
         stop 1
      end if

   contains

      subroutine move_dbl(dbl)
         real(dp) :: dbl
         i = i+1
         select case (op)
            case (extra_info_get)
            dbl = s% extra_work(i)
            case (extra_info_put)
            s% extra_work(i) = dbl
         end select
      end subroutine move_dbl

      subroutine move_int(int)
         integer :: int
         i = i+1
         select case (op)
            case (extra_info_get)
            int = s% extra_iwork(i)
            case (extra_info_put)
            s% extra_iwork(i) = int
         end select
      end subroutine move_int

      subroutine move_flg(flg)
         logical :: flg
         i = i+1
         select case (op)
            case (extra_info_get)
            flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
            if (flg) then
               s% extra_iwork(i) = 1
            else
               s% extra_iwork(i) = 0
            end if
         end select
      end subroutine move_flg

   end subroutine move_extra_info

end module run_star_extras